-- Using AceAddon-3.0
-- Creating an addon object
-- Including AceConsole functionality; "AceConsole-3.0"
-- Including AceEvent functionality; "AceEvent-3.0"
-- Including AceHook functionality; "AceHook-3.0"
ExPack = LibStub("AceAddon-3.0"):NewAddon("ExPack", "AceEvent-3.0", "AceConsole-3.0", "AceHook-3.0");
-- Using translations
local L = LibStub("AceLocale-3.0"):GetLocale("ExPack", "enUS", true);
-- Registering the options
local AceConfig = LibStub("AceConfigDialog-3.0");
local myAddonName = "ExPack";

-- Set up DataBroker for minimap button
ExPackLDB = LibStub("LibDataBroker-1.1"):NewDataObject("ExPack", {
	type = "data source",
	text = "Ghordian ExPack",
	label = "ExPack by |cff69CCF0Ghordian",
	icon = "Interface\\AddOns\\Ghordian_ExPack\\logo",
	OnClick = function(self, button)
	if button == "LeftButton" then
		ExPack:OpenOptions()
	elseif button == "RightButton" then
		if IsShiftKeyDown() then
			--ExPack.db.profile.visual.ShowBorder = not ExPack.db.profile.visual.ShowBorder;
			ExPack:ToggleVisualGoldenCrown()
		else
			--ExPack.db.profile.visual.ShowCaption = not ExPack.db.profile.visual.ShowCaption;
			ExPack:ToggleVisualCaption();
		end
	end
	end,
	OnTooltipShow = function(tooltip)
		tooltip:AddLine("ExPack by |cff69CCF0Ghordian")
		tooltip:AddLine(" ")
		tooltip:AddLine(L.minimapHint, 0.2, 1, 0.2, 1)
		tooltip:AddLine("To toggle minimap button, type '/ep minimap'")
	end
})

LibDBIcon = LibStub("LibDBIcon-1.0")

ExPack.guildBank = {}
ExPack._bagUpdateDebounce = nil
ExPack.version = "v0.3.9";
ExPack.merchantOpen = false
ExPack.enabled = true

local hooks = {}
function ExPack:RegisterAddonHook(addon, callback)
	if IsAddOnLoaded(addon) then
		callback()
	else
		hooks[addon] = callback
	end
end

-- key bindings
BINDING_HEADER_EXPACK = L["Item Expansion Pack"]
BINDING_NAME_EXPACK_CAPTION = L["Expack Show/Hide Caption"]
BINDING_NAME_EXPACK_GOLDENCROWN = L["Expack Show/Hide Goldencrown"]
BINDING_NAME_EXPACK_BANK_ONLY = L["Expack Toggle Banks Only filter"]
BINDING_NAME_EXPACK_ANY_ITEM = L["Expack Toggle Any Item filter"]
BINDING_NAME_EXPACK_ANY_EXPAN = L["Expack Toggle Any Expansion filter"]

function ExPack:ADDON_LOADED(event, addon)
	ExPack:dPrint("ADDON_LOADED; "..addon)
	if hooks[addon] then
		hooks[addon]()
		hooks[addon] = nil
	end
end

function ExPack:ToggleMinimapButton()
	self.db.profile.minimap.hide = not self.db.profile.minimap.hide;
	ExPack:UpdateMinimapButton()
end

function ExPack:UpdateMinimapButton()
	if (self.db.profile.minimap.hide) then
		LibDBIcon:Hide("ExPack")
	else
		LibDBIcon:Show("ExPack")
	end
end

function ExPack:OnInitialize()
	-- Code that you want to run when the addon is first loaded goes here.
	--ExPack:dPrint("Begin ExPack")

	-- Using AceConsole for slash commands
	--ExPack:RegisterChatCommand( "expack", "ChatCommand" );
	--ExPack:dPrint("AceConsole")
	ExPack:RegisterChatCommand( "ep", "ChatCommand" );

	--Using AceDB-3.0
	-- SavedVariables already prepared in toc file
	-- Initializing AceDB
	-- CVAR_UPDATE "autoLootDefaut", "1"
	--ExPack:dPrint("AceDB")

	if not Ghordian_ExPackDB then
		Ghordian_ExPackDB = {}
	end

	self.db = LibStub("AceDB-3.0"):New("Ghordian_ExPackDB", ExPack.defaults)
	-- Working with profiles
	if (self.db.profile == nil) or not (self.db.profile) then 
		self.db.profile = ExPack.defaults.profile; 
		ExPack.profile = self.db.profile;
	end;

	-- Standard minimap button
	LibDBIcon:Register("ExPack", ExPackLDB, self.db.profile.minimap)
	ExPack:UpdateMinimapButton()
	
	--[[ Registering the options REMOVED V3.4
	--ExPack:Print("AceConfig")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ExPack", ExPack.options);
	AceConfig:AddToBlizOptions("ExPack", "ExPack");
	]]--

	-- Standard hooking
	--ExPack:dPrint("AceHook")
	--[[ SIMPLIFIED BY ContainerFrame_GenerateFrame
	ExPack:SecureHook("OpenAllBags", "ExPack_OpenAllBags");
	ExPack:SecureHook("ToggleAllBags", "ExPack_ToggleAllBags");
	ExPack:SecureHook("ToggleBag", "ExPack_ToggleBag");
	ExPack:SecureHook("OpenBag", "ExPack_OpenBag");
	]]--
	--[[ Events:
		BAG_UPDATE
		ITEM_LOCK_CHANGED
		ITEM_UNLOCKED
		BAG_UPDATE
		BAG_UPDATE_DELAYED
	--
		https://wowpedia.fandom.com/wiki/API_C_Item.RequestLoadItemData
		C_Item.RequestLoadItemData(itemLocation)
			RequestLoadItemData
			itemLocation
			ItemLocationMixin

		C_Item.RequestLoadItemDataByID(itemInfo)
			RequestLoadItemDataByID
		itemInfo
			string : ItemLink or ID

		https://wowpedia.fandom.com/wiki/ITEM_DATA_LOAD_RESULT
		ITEM_DATA_LOAD_RESULT: itemID, success

		ItemButton.UpdateItemContextMatching

	]]--
	ExPack:SecureHook(C_Container, "PickupContainerItem", "ExPack_PickupContainerItem");

	--C_Container.SetItemSearch(searchString)
	--TODO : Filtering by expansion name/abbreviation
	--ExPack:SecureHook(C_Container, "SetItemSearch", "ExPack_SetItemSearch");

	ExPack:SecureHook("BankFrameItemButton_Update", "ExPack_BankFrameItemButton_Update");

	if _G.ContainerFrame_Update then
		ExPack:SecureHook("ContainerFrame_Update", "ExPack_ContainerFrame_Update");
	else
		--DragonFlight improvement
		ExPack:SecureHook("ContainerFrame_GenerateFrame", ExPack_ContainerFrame_GenerateFrame)
	end

	--ExPack:SecureHook(C_MerchantFrame, "MerchantFrame_Update", ExPack_MerchantFrame_Update)
	ExPack:SecureHook("MerchantFrame_Update", ExPack_MerchantFrame_Update)

	--ExPack:SecureHook("OpenMail_GetItemCounts", ExPack_OpenMail_GetItemCounts);
	--ExPack:SecureHook("OpenMailFrame_UpdateButtonPositions", ExPack_OpenMailFrame_UpdateButtonPositions);

	--Bagnon:
	if Bagnon and BagnonFrameInventory then
		BagnonFrameInventory:HookScript('onShow', function()
			ExPack:ExPack_BagnonFrameInventory()
		end)
		ExPack:dPrint("BagnonFrameInventory hooked")
	end

	do
		local function bagbrother_button(button)
			local BagID = button:GetBag();
			local slotID = button:GetID();
			local thisContainer;
			ExPack:dPrint("bagbrother_button; "..BagID.."; "..slotID);
			CleanBagFrame(BagID, slotID, button);
			--if not self.db.Enabled then
				--return
			--end
			local bag = button:GetBag();
			if type(bag) ~= "number" then
				-- try to fall back on item links, mostly for void storage which would be "vault" here
				local itemLink = button:GetItem()
				if itemLink then
					local item = Item:CreateFromItemLink(itemLink)
					-- todo - UpdateButtonFromItem(button, item, "bags")
				else
					thisContainer = ContainerTypes[ExPack.Types.BAGS];
				end
				return
			end
			-- to-do - UpdateContainerButton(button, bag)
			ExPack:ScanBagFrame(BagID, slotID, button, thisContainer)
		end
		ExPack:RegisterAddonHook("Bagnon", function()
			-- todo - hooksecurefunc(Bagnon.Item, "Update", bagbrother_button)
			--ExPack:SecureHook(Bagnon.Item, "Update", bagbrother_button)
			hooksecurefunc(Bagnon.Item, "Update", bagbrother_button);
			ExPack:dPrint("register; bagbrother_button; ");
		end)
		ExPack:dPrint("hook; bagbrother_button; ");
	end

	--ExPack:dPrint("AceEvent")
	if (ExPack:CheckDB() == true) then
		-- Including AceEvent functionality
		ExPack:RegEvents();
	end;

	-- Using AceConsole for output; used as a mixin for AceAddon
	ExPack:Print("ExPack by |cff69CCF0Ghordian");
end

function ExPack:OnEnable()
	ExPack:Print("OnEnable |cff00CC00LOADED");
	-- Called when the addon is enabled
	ExPack:RegEvents();
	--
	ExPack:CreateConfigPanels();
end

function ExPack:OnDisable()
	ExPack:Print("OnDisable |cff00CC00LOADED");
	-- Called when the addon is disabled
	ExPack:UnRegEvents();
	-- todo - finally cleanup containers
end

function ExPack:ChatCommand(input)
	-- Process the slash command ('input' contains whatever follows the slash command)
	if (input == "minimap") then
		ExPack:ToggleMinimapButton()
	else
		ExPack:OpenOptions()
	end
end;

function ExPack:OpenOptions()
	-- first time opens Interface Frame
	InterfaceOptionsFrame_OpenToCategory("ExPack");
	-- Second time opens Category Option Panel
	InterfaceOptionsFrame_OpenToCategory("ExPack");
end;

function ExPack:CheckDB()
	if self.db.profile.advanced.UseLibItemInfo then 	
		local LII = LibStub("LibItemInfo-1.0");
		if LII == nil then
			if self.db.profile.reports.ErrorReporting then
				ExPack:Print("LibItemInfo |cffCC0000NOT LOADED");
			else
				ExPack:Print("LibItemInfo |cff00CC00LOADED");
			end
			return false;
		end
	elseif self.db.profile.advanced.UseItemVersion then
		if not (ExPack.itemIdToVersionId) then return false; end;
		if not (ExPack.versionIdToVersion) then return false; end;
		ExPack:Print("ItemVersion |cffCC0000NOT LOADED");
	end
	return true;
end;

function ExPack:RegEvents()
	-- Subscribing to events
	--ExPack:dPrint("RegEvents")

	ExPack:RegisterEvent("ADDON_LOADED")

	-- Events for Guild Bank UI Dragonflight 10.0.02
	--REMOVED ExPack:RegisterEvent("GUILDBANKFRAME_OPENED")
	--REMOVED ExPack:RegisterEvent("GUILDBANKFRAME_CLOSED") 
	ExPack:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")

	-- Events for Player bank UI
	ExPack:RegisterEvent("BANKFRAME_OPENED")
	ExPack:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")

	ExPack:RegisterEvent("BAG_UPDATE");
	ExPack:RegisterEvent("BAG_UPDATE_DELAYED");

	ExPack:RegisterEvent("MERCHANT_SHOW");
	ExPack:RegisterEvent("MERCHANT_CLOSED");
	ExPack:RegisterEvent("MERCHANT_FILTER_ITEM_UPDATE");
	ExPack:RegisterEvent("MERCHANT_UPDATE");

	ExPack:RegisterEvent("LOOT_READY");
	ExPack:RegisterEvent("LOOT_OPENED");
	ExPack:RegisterEvent("LOOT_CLOSED");
	ExPack:RegisterEvent("LOOT_SLOT_CLEARED");

	-- https://wowpedia.fandom.com/wiki/Events - /etrace

end

function ExPack:UnRegEvents()

	-- todo - clear previous ExPack info for all the containers

	-- Events for Guild Bank UI Dragonflight 10.0.02
	--REMOVED ExPack:UnregisterEvent("GUILDBANKFRAME_OPENED")
	ExPack:UnregisterEvent("GUILDBANKBAGSLOTS_CHANGED")

	ExPack:UnregisterEvent("BAG_UPDATE");
	ExPack:UnregisterEvent("BAG_UPDATE_DELAYED");

	ExPack:UnregisterEvent("MERCHANT_SHOW");
	ExPack:UnregisterEvent("MERCHANT_CLOSED");
	ExPack:UnregisterEvent("MERCHANT_FILTER_ITEM_UPDATE");
	ExPack:UnregisterEvent("MERCHANT_UPDATE");

	ExPack:UnregisterEvent("LOOT_READY");
	ExPack:UnregisterEvent("LOOT_OPENED");
	ExPack:UnregisterEvent("LOOT_CLOSED");
	ExPack:UnregisterEvent("LOOT_SLOT_CLEARED");

	-- Events for Player bank UI
	ExPack:UnregisterEvent("BANKFRAME_OPENED")
	ExPack:UnregisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")

end

--[[
function ExPack:GUILDBANKFRAME_OPENED()
	--if not self.db.profile.general.Enabled then return end
	ExPack:dPrint("OnGuildBankFrameOpened")

	ExPack.guildBank.isGuildBankOpen = true
	ExPack.guildBank.numTabs = GetNumGuildBankTabs()
	ExPack.guildBank.guildFunds = GetGuildBankMoney()
	ExPack.guildBank.tab = GetCurrentGuildBankTab()

	--register the _CLOSED event whenever the bank is opened
	--REMOVED ExPack:RegisterEvent("GUILDBANKFRAME_CLOSED") 
	--REDO ExPack:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
end

function ExPack:GUILDBANKFRAME_CLOSED() 
	--if not self.db.profile.general.Enabled then return end
	ExPack:dPrint("OnGuildBankFrameClosed")

	--REMOVED ExPack:UnregisterEvent("GUILDBANKFRAME_CLOSED") 
	ExPack:UnregisterEvent("GUILDBANKBAGSLOTS_CHANGED")

	ExPack:GuildBankFrame_RemoveExPack();
end
]]--

function ExPack:GUILDBANKBAGSLOTS_CHANGED() 
	ExPack.guildBank.isGuildBankOpen = true
	--if not self.db.profile.general.Enabled then return end
	--ExPack:dPrint("GUILDBANKBAGSLOTS_CHANGED");
	ExPack_Debounce(ExPack_OnGuildBankFrameChanged, 0.01)
end

--[[
	BagnonGuildFrame1 name found thanks to 
	DEBUGGING; scanned via /tinspect GlobalName
]]--

function ExPack_OnGuildBankFrameChanged() 
	if not ExPack.db.profile.general.Enabled then return end
	if ExPack.guildBank.isGuildBankVisited == false then
		ExPack.guildBank.isGuildBankVisited = true
	end
	ExPack:dPrint("OnGuildBankFrameChanged")
	local num, itemString
	if Bagnon or BagBrother then
		ExPack:dPrint("BagnonGuildFrame1; ")
		ExPack:vardump(BagnonGuildFrame1)
	end
	if (GuildBankFrame and GuildBankFrame:IsVisible())
		or (BagnonGuildFrame1 and BagnonGuildFrame1:IsVisible()) then
		local tab = GetCurrentGuildBankTab();
		if ( ExPack.guildBank.tab ~= tab ) then
			if (tab ~= nil) and (ExPack.guildBank.tab ~= nil) then
				--ExPack:jPrint("GuildBankFrame.Tab differ; "..tab.."; "..ExPack.guildBank.tab);
			end
			ExPack:GuildBankFrame_RemoveExPack();
		end
		ExPack.guildBank.tab = tab;

		-- GuildBankFrame_OnLoad(..)
		-- GuildBankFrame.mode == "bank"
		local button, index, column;
		local NumSlotsPerGuildBankGroup = 14 --NUM_SLOTS_PER_GUILDBANK_GROUP;
		local btnCount = 0
		local thisContainer = ContainerTypes[ExPack.Types.GUILDBANK]
		local size = thisContainer:GetSize();
		ExPack:dPrint("GuildBankSize; "..size.."; "..NumSlotsPerGuildBankGroup)
		for slot=1, size do
			index = mod(slot, NumSlotsPerGuildBankGroup);
			if ( index == 0 ) then
				index = NumSlotsPerGuildBankGroup;
			end
			column = ceil((slot-0.5) / NumSlotsPerGuildBankGroup);
			--ExPack:jPrint("GuildBankColumn; tab; "..tab.."; slot; "..slot.."; column"..column.."; index; "..index);
			button = _G["GuildBankColumn"..column.."Button"..index];
			if button == nil then
			--	ExPack:dPrint("GuildBankColumn fail")
				if (GuildBankFrame) then
					button = GuildBankFrame.Columns[column].Buttons[index];
				elseif (BagnonGuildFrame1) then
					-- todo button = BagnonGuildFrame1.Buttons[index];
				end
			end
			if button ~= nil then
				ExPack:ScanBagFrame(tab, slot, button, thisContainer);
				btnCount = btnCount + 1
			else
				ExPack:dPrint("GuildBankFrame fail")
			end
		end
		if ( btnCount < size ) then
			ExPack:jPrint("GuildBankColumn; "..btnCount.."; "..size);
		end
		ExPack:dReset();
	else
		ExPack:dPrint("No GuildBankFrame; ");
	end
	ExPack.guildBank.isGuildBankOpen = false
end

function ExPack:BANKFRAME_OPENED()
	if not self.db.profile.general.Enabled then return end
	ExPack:dPrint("OnBankFrameOpened");

	local frame = BankFrame;
	if (frame.activeTabIndex == 1) then
		for bagID = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do		-- 5 to 11
			ExPack:ScanBag(bagID)
		end
		--ExPack:ScanBank();
	elseif (frame.activeTabIndex == 2) then
		ExPack:ScanReagentBank();
	end

	ExPack:RegisterEvent("BANKFRAME_CLOSED")
	ExPack:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
	ExPack:RegisterEvent("REAGENTBANK_UPDATE")
	ExPack:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")
end

function ExPack:BANKFRAME_CLOSED()
	if not self.db.profile.general.Enabled then return end
	--
	ExPack:dPrint("OnBankFrameClosed")
	
	ExPack:UnregisterEvent("BANKFRAME_CLOSED")
	ExPack:UnregisterEvent("PLAYERBANKSLOTS_CHANGED")

	-- done ; clean after close
	ExPack:BankFrame_RemoveExPack();
end

function ExPack:PLAYERBANKSLOTS_CHANGED(slot)
	if not self.db.profile.general.Enabled then return end
	--
	ExPack:dPrint("OnPlayerBankSlotsChanged; "..slot)
	ExPack:dReset()

	ExPack:ScanContainer(BANK_CONTAINER, ExPack.Types.BANK)
end

function ExPack:REAGENTBANK_UPDATE()
	if not self.db.profile.general.Enabled then return end
	ExPack:dPrint("ReagentbankUpdate")

	ExPack:ScanReagentBank()
end

function ExPack:PLAYERREAGENTBANKSLOTS_CHANGED(slot)
	if not self.db.profile.general.Enabled then return end
	ExPack:dPrint("OnReagentBankFrameChanged; "..slot)

	ExPack:ScanReagentBank()
end

function ExPack:ITEM_LOCK_CHANGED(self, event, bagOrSlotIndex, slotIndex)
	local bag = bagOrSlotIndex or -1
	local slot = slotIndex or -1
	ExPack:dPrint("ITEM_LOCK_CHANGED; "..bag.."; "..slot)
	--ExPack:CleanPickupItem()
end

function ExPack:MERCHANT_SHOW()
	ExPack:dPrint("MERCHANT_SHOW; ")
	ExPack.merchantOpen = true
end	
	
function ExPack:MERCHANT_CLOSED()
	ExPack:dPrint("MERCHANT_CLOSED; ")
	if ExPack.merchantOpen then
		ExPack_merchantOpen = false
		ExPack_ResetVendorPage("close")
	end
end

function ExPack:MERCHANT_FILTER_ITEM_UPDATE()
	ExPack:dPrint("MERCHANT_FILTER_ITEM_UPDATE; ")
	if ExPack.merchantOpen then
		ExPack_MerchantFrame_Update()
	end
end

function ExPack:MERCHANT_UPDATE()
	ExPack:dPrint("MERCHANT_UPDATE; ")
	if ExPack.merchantOpen then
		ExPack_MerchantFrame_Update()
	end
end

-- C_Loot - loot mechanism : begin

-- https://wowpedia.fandom.com/wiki/LOOT_READY
function ExPack:LOOT_READY(event)
	-- numLootItems = GetNumLootItems()
	ExPack:dPrint("Expack; "..event)
	ExPack.merchantOpen = true
end

-- https://wowpedia.fandom.com/wiki/LOOT_OPENED
	--autoLoot:boolean - Equal to CVar autoLootDefault.
	--isFromItem:boolean
function ExPack:LOOT_OPENED(event, autoLoot, isFromItem)
	ExPack:dPrint("Expack; "..event)
end

-- https://wowpedia.fandom.com/wiki/LOOT_CLOSED
function ExPack:LOOT_CLOSED(event)
	ExPack:dPrint("Expack; "..event)
	if ExPack.merchantOpen then
		ExPack_merchantOpen = false
		--ExPack_ResetVendorPage("close")
	end
end

function ExPack:BAG_UPDATED_COOLDOWN(event)
	ExPack:dPrint("Expack; "..event)
end

--  https://wowpedia.fandom.com/wiki/LOOT_SLOT_CHANGED
-- lootSlot:number
function ExPack:LOOT_SLOT_CHANGED(event, lootSlot)
	ExPack:dPrint("Expack; "..event)
end

-- https://wowpedia.fandom.com/wiki/LOOT_SLOT_CLEARED
--  lootSlot:number
function ExPack:LOOT_SLOT_CLEARED(event, lootSlot)
	ExPack:dPrint("Expack; "..event)
end

-- loot mechanism : end

function ExPack_ResetVendorPage(info)
	ExPack:dPrint("ExPack_ResetVendorPage; "..info)
	for i = 1, MERCHANT_ITEMS_PER_PAGE, 1 do
		local btnName = "MerchantItem" .. i .. "ItemButton";
		local btnIconName = btnName.."ExPack";
		local itemButton = _G[btnName]
		if (itemButton) then
			if (itemButton.ExPack) then
				itemButton.ExPack:Hide()
				itemButton.ExPack = null
			end
		end

		local tframe = _G[btnIconName]
		if (tframe) then
			tframe:Hide()
			--_G[btnIconName] = null
		end
	end
end

function ExPack_MerchantFrame_Update()
	ExPack:dPrint("ExPack_MerchantFrame_Update; "..MerchantFrame.page)
	if ((ExPack.merchantOpen == false) or (ExPack.enabled == false)) then
		ExPack:dPrint("MerchantFrame_Update; loot/mail?")
		return ;
	end

	ExPack_ResetVendorPage("update; page;"..MerchantFrame.page);

	local profile = ExPack.db.profile

	if not profile.filters.ShowMerchants then
		return
	end

	for i = 1, MERCHANT_ITEMS_PER_PAGE, 1 do
		local index = (((MerchantFrame.page - 1) * MERCHANT_ITEMS_PER_PAGE) + i);
		local itemLink = GetMerchantItemLink(index);
		local itemID = nil
		if (itemLink) then
			--ExPack:dPrint("MerchantFrame_Update; ItemLink found; "..index)
			--SharedXML\LinkUtil
			itemID = GetItemInfoFromHyperlink(itemLink)
			if (itemID) then
				--ExPack:dPrint("MerchantFrame_Update; ItemID found; "..itemID)
			end
		end

		local btnName = "MerchantItem" .. i .. "ItemButton";
		local itemButton = _G[btnName]

		if (itemID) then
			local tframe = ExPack_GetExPackFrame(itemButton, itemID, itemLink, 0, 0)

			if (tframe) then
				if (profile.filters.ShowMerchants) then
					ExPack:dPrint("MerchantFrame_Update; tframe; Show; "..index.."; "..itemID.."; "..btnName)
					if profile.visual.ShowBorder then
						tframe:Show();
						ExPack:dPrint("MerchantFrame_Update; tframe; ShowCrow; "..itemID)
					end
					if profile.visual.ShowCaption then
						itemButton.ExPack:Show();
						ExPack:dPrint("MerchantFrame_Update; itemButton; ShowCaption; "..itemID)
					end
				else
					ExPack:dPrint("MerchantFrame_Update; Hide; "..index.."; "..btnName)
					tframe:Hide();
					itemButton.ExPack:Hide();
				end
			end
		end
	end
end

function ExPack_GetExpackInfoFromItemID(itemID , expacID)

	local expname = "n/a";
	local clr = "|cFFffff00";

	if (itemID > 0) and (expacID == 0) then
		local itemVersion, itemExpac = ExPack:myGetVersion(itemID);
		if itemVersion then
			--ExPack:vardump (itemVersion);
			itemExpac = ExPack:myGetVersionExpac(itemVersion);
			-- if itemExpac ~= nil then
				-- expname = itemExpac.shortName;
				-- clr = itemExpac.color;
				-- expacID = itemVersion.major;
			-- else
				-- ExPack:dPrint("itemExpac nil; "..itemID.." ; "..quality)
				-- if not self.db.profile.advanced.UseItemVersion then
					-- don't show "n/a"
					-- expname = "";
				-- end
			-- end
		end

		if itemExpac then
			if itemExpac ~= nil then
				expname = itemExpac.shortName;
				clr = itemExpac.color;
				expacID = itemExpac.expansionID;
			else
				ExPack:dPrint("itemExpac nil; "..itemID.." ; "..quality)
				if not self.db.profile.advanced.UseItemVersion then
					-- don't show "n/a"
					expname = "";
				end
			end
		else
			quality = quality or 0;
			--ExPack:dPrint("itemVersion nil; "..itemID.." ; "..quality)
			if not ExPack.db.profile.advanced.UseItemVersion then
				-- don't show "n/a"
				expname = "";
			end
		end
	end

	return expname, clr, expacID;
end

function ExPack_GetColor(clr)
	local Color = {}
	local r,g,b
	r, g, b = clr:match("|cFF(%x%x)(%x%x)(%x%x)")
	r, g, b = tonumber(r, 16), tonumber(g, 16), tonumber(b, 16)
	Color.r, Color.g, Color.b = r / 255, g / 255, b / 255
	return Color
end

function ExPack_GetExPackFrame(bagItemButton, itemID, itemLink, bankCol, bankIdx)
	if (bagItemButton == nil) then
		return
	end
	local bagItemButtonName = bagItemButton:GetName()
	if (bagItemButtonName == nil) then
		bagItemButtonName = "GuildBankItemButtonC" .. bankCol .. "B" .. bankIdx
	end

	local expname, clr, expacID  = ExPack_GetExpackInfoFromItemID(itemID, 0);
	local bCanShow = ExPack:CanShowItemLink(itemLink, expacID);

	local Color = ExPack_GetColor(clr);

	local btnName = bagItemButtonName .. "ExPack";
	local tframe = _G[btnName]

	local profile = ExPack.db.profile

	if profile.visual.ShowBorder and bCanShow then
		if (tframe == nil) then
			-- ShowBorder/Golden Crown
			tframe = CreateFrame("Frame", btnName, bagItemButton)
			tframe:ClearAllPoints()
			tframe:SetWidth(16)
			tframe:SetHeight(16)
			tframe:SetPoint("BOTTOMLEFT", bagItemButton, "BOTTOMLEFT", -4, -4)
			tframe.icon = tframe:CreateTexture()
			tframe.icon:SetTexture("Interface/Addons/Ghordian_ExPack/images/GoldenCrown.tga")
			tframe.icon:SetAllPoints(tframe)
			tframe.icon:SetBlendMode("BLEND")
			tframe.icon:SetPoint("CENTER", tframe)
		end
		if (itemID > 0) then
			tframe.icon:SetVertexColor(Color.r, Color.g, Color.b, 1)
			--tframe:Show();
			ExPack:dPrint("MerchantFrame_Update; tframe; ShowCrow; "..itemID)
		else
			tframe:Hide();
			ExPack:dPrint("MerchantFrame_Update; tframe; No ItemID; "..itemID)
		end
	else
		if (tframe) then
			tframe:Hide();
			--tframe.icon = null;
			ExPack:dPrint("MerchantFrame_Update; tframe; HideCrow; "..itemID)
		end
	end

	-- ShowCaption
	if profile.visual.ShowCaption and bCanShow then
		if not bagItemButton.ExPack then
			bagItemButton.ExPack = bagItemButton:CreateFontString(nil, "OVERLAY", "NumberFontNormalSmall")
			bagItemButton.ExPack:SetPoint("TOPLEFT", bagItemButton, 0, -2)
			bagItemButton.ExPack:SetTextColor(1,1,0) -- yellow!
			bagItemButton.ExPack:SetText("")
		end
		if (itemID > 0) then
			-- Apply appearance configuration
			bagItemButton.ExPack:ClearAllPoints()
			bagItemButton.ExPack:SetPoint(profile.visual.Position, unpack(ExPack.PositionOffsets[profile.visual.Position]))
			bagItemButton.ExPack:SetFontObject(ExPack.Fonts[profile.visual.Font] or NumberFontNormal)
			bagItemButton.ExPack:SetTextColor(Color.r, Color.g, Color.b)
			bagItemButton.ExPack:SetText( expname )
			bagItemButton.ExPack:Hide()
			--bagItemButton.ExPack:Show()
			--ExPack:dPrint("MerchantFrame_Update; tframe; ShowCaption; "..itemID.."; "..expname)
		else
			bagItemButton.ExPack:SetText( "" )
		end
	else
		if bagItemButton.ExPack then
			bagItemButton.ExPack:Hide()
		end
		--ExPack:dPrint("MerchantFrame_Update; tframe; HideCaption; "..itemID)
	end

	return tframe
end

local item_update_time = 0;

function ExPack:LOOT_SLOT_CLEARED(event)
	item_update_time = GetTime() - item_update_time
	ExPack:dPrint("LOOT_SLOT_CLEARED; "..item_update_time.."; "..event)
	--if not self.db.profile.general.Enabled then return end
	--	ExPack:ExPack_BagsUpdate(event, 0);
	item_update_time = GetTime();
end

function ExPack:ITEM_LOCK_CHANGED(event)
	item_update_time = GetTime() - item_update_time
	ExPack:dPrint("ITEM_LOCK_CHANGED; "..item_update_time.."; "..event)
	--if not self.db.profile.general.Enabled then return end
	--	ExPack:ExPack_BagsUpdate(event, 0);
	item_update_time = GetTime();
end

function ExPack:ITEM_LOCKED(event)
	item_update_time = GetTime() - item_update_time
	ExPack:dPrint("ITEM_LOCKED; "..item_update_time.."; "..event)
	--if not self.db.profile.general.Enabled then return end
	--	ExPack:ExPack_BagsUpdate(event, 0);
	item_update_time = GetTime();
end

local bag_update_time = 0;

--	https://wowpedia.fandom.com/wiki/BAG_UPDATE_DELAYED
function ExPack:BAG_UPDATE_DELAYED(event)
	bag_update_time = GetTime() - bag_update_time
	ExPack:dPrint("ExPack; "..event.."; "..bag_update_time)
	--if not self.db.profile.general.Enabled then return end
	--
	ExPack:ExPack_BagsUpdate(event, 0);
	bag_update_time = GetTime();
end

--https://wowpedia.fandom.com/wiki/BAG_UPDATE
--	BAG_UPDATE: bagID
function ExPack:BAG_UPDATE(event, bagID)
	bag_update_time = GetTime() - bag_update_time
	ExPack:dPrint("ExPack; "..event.."; "..bag_update_time.."; "..bagID)
	--if not self.db.profile.general.Enabled then return end
	--
	--if (bag_update_time > 1) then
		ExPack:ExPack_BagsUpdate(event, bagID);
	--end
	bag_update_time = GetTime();
end

function ExPack_OnEvent(self, event, ...)

  --  if ((event == "BAG_UPDATE") or (event == "BAG_CONTAINER_UPDATE")) then
  --      ExPack_Debounce(ExPack_ExecuteButtonUpdates, 0.01)
  --  elseif (event == "GUILDBANKBAGSLOTS_CHANGED") then
  --      ExPack_Debounce(ExPack_ExecuteButtonUpdatesGuildBank, 0.01)
  --  else
  --      --print(self, event, ...)
		ExPack:dPrint("ExPack_OnEvent; "..event);
  --  end

end

--https://wowpedia.fandom.com/wiki/INVENTORY_SEARCH_UPDATE
function ExPack:INVENTORY_SEARCH_UPDATE(event, ...)

	ExPack:dPrint("ExPack; "..event);

end

function ExPack_Debounce(fn, seconds)
	if (ExPack._bagUpdateDebounce) then
		ExPack._bagUpdateDebounce.cancelled = true
	end

	local timer = {
		func = fn,
		cancelled = false,
	}

	timer.callback = function()
		if not timer.cancelled then
			timer.func()
			ExPack._bagUpdateDebounce = nil
		end
	end

	ExPack._bagUpdateDebounce = timer
	C_Timer.After(seconds, timer.callback)
		gTimer = timer
end

--[[

function ExPack_OnLoad(self, event)
	ExPack:dPrint("ExPack_OnLoad; ");
end

function ExPack_ExecuteButtonUpdates()
	ExPack:dPrint("ExPack_ExecuteButtonUpdates; ");
end

function ExPack_ExecuteButtonUpdatesGuildBank()
	ExPack:dPrint("ExPack_ExecuteButtonUpdatesGuildBank; ");
end

function ExPack_CloseGuildBankFrame()
	ExPack:dPrint("ExPack_CloseGuildBankFrame; ");
	ExPack:BANKFRAME_CLOSED()
end

--]]

-- key bindings
function ExPack:ToggleFrame()
	ExPack:dPrint(BINDING_HEADER_EXPACK);
	ExPack:OpenOptions()
end

function ExPack:ToggleVisualCaption()
	ExPack:dPrint(BINDING_NAME_EXPACK_CAPTION );
	ExPack.db.profile.visual.ShowCaption = not ExPack.db.profile.visual.ShowCaption;
	ExPack:ExPack_ScanBags();
end

function ExPack:ToggleVisualGoldenCrown()
	ExPack:dPrint(BINDING_NAME_EXPACK_GOLDENCROWN );
	ExPack.db.profile.visual.ShowBorder = not ExPack.db.profile.visual.ShowBorder;
	ExPack:ExPack_ScanBags();
end

function ExPack:ToggleFilterBanksOnly()
	ExPack:dPrint(BINDING_NAME_EXPACK_BANK_ONLY );
	ExPack.db.profile.filters.ShowBanks = not ExPack.db.profile.filters.ShowBanks;
	ExPack:ExPack_ScanBags();
end

function ExPack:ToggleFilterAnyItem()
	ExPack:dPrint(BINDING_NAME_EXPACK_ANY_ITEM );
	ExPack.db.profile.filters.ShowAll = not ExPack.db.profile.filters.ShowAll;
	ExPack:ExPack_ScanBags();
end

function ExPack:ToggleFilterAnyExpansion()
	ExPack:dPrint(BINDING_NAME_EXPACK_ANY_EXPAN );
	ExPack.db.profile.filters.AnyExpansion = not ExPack.db.profile.filters.AnyExpansion;
	ExPack:ExPack_ScanBags();
end

--[[

function ExPack_GetInboxItem(messageIndex, attachmentIndex)
	-- GetInboxItem(InboxFrame.openMailID, i);
	ExPack:dPrint( "GetInboxItem; "..messageIndex..";"..attachmentIndex );
end

function ExPack_OpenMail_GetItemCounts(isTakeable, textCreated, money)

	ExPack:dPrint( "OpenMail_GetItemCounts; "..BoolToStr(isTakeable).."; "..BoolToStr(textCreated).."; "..money );

end

function ExPack_OpenMailFrame_UpdateButtonPositions()

	ExPack:dPrint( "OpenMailFrame_UpdateButtonPositions; " );

	for i=1, ATTACHMENTS_MAX_RECEIVE do
		local attachmentButton = OpenMailFrame.OpenMailAttachments[i];
		if HasInboxItem(InboxFrame.openMailID, i) then

			local name, itemID, itemTexture, count, quality, canUse = GetInboxItem(InboxFrame.openMailID, i);
			if name then
				ExPack:dPrint( "attachmentButton; "..i.."; "..itemID.."; "..count );
			end

			local itemLink = GetInboxItemLink(InboxFrame.openMailID, i);
			if (itemLink) then
				ExPack:dPrint( "attachmentButton; "..i.."; itemLink" );
			end

			if (itemID) then
				local tframe = ExPack_GetExPackFrame(attachmentButton, itemID, itemLink, 0, 0);

				if (tframe) then
				--if (profile.filters.ShowAttachments) then
				--	ExPack:dPrint("Attachments_Update; tframe; Show; "..index.."; "..itemID.."; "..btnName)
				--	tframe:Show()
				--else
				--	ExPack:dPrint("Attachments_Update; tframe; Hide; "..index.."; "..btnName)
				--	tframe:Hide()
				--end
				end
			end

		end
	end

end

--]]
